from __future__ import annotations

import argparse
import json
import pathlib
import sys
from collections.abc import Mapping

from . import combiner, exceptions, toml


def get_argument_parser(
    dimensions: Mapping[str, list[str]] | None,
) -> argparse.ArgumentParser:
    """Get the command-line argument parser."""
    arg_parser = argparse.ArgumentParser(
        description="Create combined configurations from a TOML file",
        add_help=(dimensions is not None),
    )
    arg_parser.add_argument(
        "config",
        type=pathlib.Path,
        help="Path to the TOML configuration file",
    )
    if dimensions:
        group = arg_parser.add_argument_group(
            "dimensions",
            "Filter the generated outputs by dimensions",
        )

        for name, values in dimensions.items():
            group.add_argument(
                f"--{name}", choices=values, help=f"Limit to given {name}"
            )

    return arg_parser


def cli(argv) -> int:
    """Main entry point."""

    # Parse the config file argument to get the dimensions
    arg_parser = get_argument_parser(dimensions=None)
    args, _ = arg_parser.parse_known_args(argv)

    try:
        dict_config = toml.loads(args.config.read_text())
    except FileNotFoundError:
        print(f"Configuration file not found: {args.config}", file=sys.stderr)
        return 1
    except exceptions.TomlDecodeError as exc:
        print(exc, file=sys.stderr)
        return 1

    try:
        config = combiner.build_config(dict_config)
    except exceptions.TomlCombineError as exc:
        print(exc, file=sys.stderr)
        return 1

    # Parse all arguments
    arg_parser = get_argument_parser(dimensions=config.dimensions)
    args = arg_parser.parse_args(argv)

    dimensions_filter = {
        key: value
        for key, value in vars(args).items()
        if key in config.dimensions and value
    }
    # Generate final configurations for each output
    try:
        result = combiner.generate_outputs(config=config, **dimensions_filter)
    except exceptions.TomlCombineError as exc:
        print(exc, file=sys.stderr)
        return 1

    if not result:
        print("No outputs found", file=sys.stderr)

    print(json.dumps(result, indent=2))

    return 0


def run_cli():
    sys.exit(cli(sys.argv[1:]))



from __future__ import annotations

import pathlib
from typing import Any, cast, overload

from . import combiner, toml


# Provide already parsed config
@overload
def combine(
    *, config: dict[str, Any], **filters: str | list[str]
) -> dict[str, Any]: ...
# Provide toml config content
@overload
def combine(*, config: str, **filters: str | list[str]) -> dict[str, Any]: ...
# Provide toml config file path
@overload
def combine(
    *, config_file: str | pathlib.Path, **filters: str | list[str]
) -> dict[str, Any]: ...


def combine(*, config=None, config_file=None, **filters):
    """
    Generate outputs of configurations based on the provided TOML
    configuration.

    Args:
        config: The TOML configuration as a string or an already parsed dictionary.
        OR:
        config_file: The path to the TOML configuration file.
        **filters: Filters to apply to the combined configuration
            (dimension="value" or dimension=["list", "of", "values"]).

    Returns:
        dict[str, Any]: The combined configuration ({"output_id": {...}}).
    """
    if (config is None) is (config_file is None):
        raise ValueError("Either 'config' or 'config_file' must be provided.")

    if isinstance(config, dict):
        dict_config = config
    else:
        if config_file:
            config_string = pathlib.Path(config_file).read_text()
        else:
            config = cast(str, config)
            config_string = config

        dict_config = toml.loads(config_string)

    config_obj = combiner.build_config(dict_config)

    return combiner.generate_outputs(config_obj, **filters)



from __future__ import annotations

import pytest
import tomli

from toml_combine import combiner, exceptions


@pytest.mark.parametrize(
    "small_override, large_override, dimensions",
    [
        pytest.param(
            {"env": "prod"},
            {"env": "prod", "region": "eu"},
            {"env": ["prod"], "region": ["eu"]},
            id="less_specific_override_comes_first",
        ),
        pytest.param(
            {"env": "prod", "region": "eu"},
            {"env": "prod", "service": "web"},
            {"env": ["prod"], "region": ["eu"], "service": ["web"]},
            id="different_dimensions_sorted_by_dimension",
        ),
        pytest.param(
            {"env": "prod"},
            {"region": "eu"},
            {"env": ["prod"], "region": ["eu"]},
            id="completely_different_dimensions",
        ),
    ],
)
def test_override_sort_key(small_override, large_override, dimensions):
    small_key = combiner.override_sort_key(
        combiner.Override(when=small_override, config={}), dimensions
    )
    large_key = combiner.override_sort_key(
        combiner.Override(when=large_override, config={}), dimensions
    )
    assert small_key < large_key


@pytest.mark.parametrize(
    "a, b, expected",
    [
        pytest.param(
            {"a": 1, "b": 2},
            {"b": 3, "c": 4},
            {"a": 1, "b": 3, "c": 4},
            id="normal_dicts",
        ),
        pytest.param(
            {"a": {"b": 1, "c": 2}},
            {"a": {"c": 3}},
            {"a": {"b": 1, "c": 3}},
            id="nested_dicts",
        ),
    ],
)
def test_merge_configs__dicts(a, b, expected):
    assert combiner.merge_configs(a, b) == expected


def test_merge_configs__dicts_error():
    with pytest.raises(ValueError):
        combiner.merge_configs({"a": 1}, {"a": {"b": 2}})


@pytest.mark.parametrize(
    "output, expected",
    [
        pytest.param(
            combiner.Output(dimensions={"env": "dev"}),
            {
                "a": 1,
                "b": 2,
                "c": 3,
                "d": {"e": {"h": {"i": {"j": 4}}}},
                "g": 6,
                "dimensions": {"env": "dev"},
            },
            id="no_matches",
        ),
        pytest.param(
            combiner.Output(dimensions={"env": "prod"}),
            {
                "a": 10,
                "b": 2,
                "c": 30,
                "d": {"e": {"h": {"i": {"j": 40}}}},
                "g": 60,
                "dimensions": {"env": "prod"},
            },
            id="single_match",
        ),
        pytest.param(
            combiner.Output(dimensions={"env": "staging"}),
            {
                "a": 1,
                "b": 200,
                "c": 300,
                "d": {"e": {"h": {"i": {"j": 400}}}},
                "f": 500,
                "g": 6,
                "dimensions": {"env": "staging"},
            },
            id="dont_override_if_match_is_more_specific",
        ),
    ],
)
def test_generate_output(output: combiner.Output, expected: dict[str, int]):
    default = {
        "a": 1,
        "b": 2,
        "c": 3,
        "d": {"e": {"h": {"i": {"j": 4}}}},
        "g": 6,
    }

    overrides = [
        combiner.Override(
            when={"env": ["prod"]},
            config={
                "a": 10,
                "c": 30,
                "d": {"e": {"h": {"i": {"j": 40}}}},
                "g": 60,
            },
        ),
        combiner.Override(
            when={"env": ["staging"]},
            config={
                "b": 200,
                "c": 300,
                "d": {"e": {"h": {"i": {"j": 400}}}},
                "f": 500,
            },
        ),
        combiner.Override(
            when={"env": ["staging"], "region": ["us"]},
            config={"f": 5000, "g": 6000},
        ),
    ]

    result = combiner.generate_output(
        output=output,
        default=default,
        overrides=overrides,
    )
    assert result == expected


@pytest.mark.parametrize(
    "output, override, expected",
    [
        (
            combiner.Output(dimensions={"env": "dev"}),
            combiner.Override(when={"env": ["dev"]}, config={}),
            True,
        ),
        (
            combiner.Output(dimensions={"env": "dev"}),
            combiner.Override(when={"env": ["staging"]}, config={}),
            False,
        ),
        (
            combiner.Output(dimensions={"env": "dev"}),
            combiner.Override(when={"env": ["dev", "staging"]}, config={}),
            True,
        ),
        (
            combiner.Output(dimensions={"env": "staging"}),
            combiner.Override(when={"region": ["us"]}, config={}),
            False,
        ),
        (
            combiner.Output(dimensions={"env": "dev"}),
            combiner.Override(when={"env": ["dev"], "region": ["us"]}, config={}),
            False,
        ),
        (
            combiner.Output(dimensions={"env": "dev", "region": "us"}),
            combiner.Override(when={"env": ["dev"]}, config={}),
            True,
        ),
    ],
)
def test_output_matches_override(output, override, expected):
    result = combiner.output_matches_override(output=output, override=override)
    assert result == expected


def test_build_config():
    raw_config = """
    [dimensions]
    env = ["dev", "staging", "prod"]

    [default]
    foo = "bar"

    [[output]]
    env = "dev"

    [[output]]
    env = ["staging", "prod"]

    [[override]]
    when.env = ["dev", "staging"]
    foo = "baz"

    [[override]]
    when.env = "prod"
    foo = "qux"
    """

    config_dict = tomli.loads(raw_config)
    config = combiner.build_config(config_dict)

    assert config == combiner.Config(
        dimensions={"env": ["dev", "staging", "prod"]},
        outputs=[
            combiner.Output(
                dimensions={"env": "dev"},
            ),
            combiner.Output(
                dimensions={"env": "staging"},
            ),
            combiner.Output(
                dimensions={"env": "prod"},
            ),
        ],
        default={"foo": "bar"},
        overrides=[
            combiner.Override(
                when={"env": ["dev", "staging"]},
                config={"foo": "baz"},
            ),
            combiner.Override(
                when={"env": ["prod"]},
                config={"foo": "qux"},
            ),
        ],
    )


def test_create_outputs__duplicate_overrides():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [templates]
    foo = "bar"

    [[output]]
    env = "dev"

    [[override]]
    when.env = "prod"
    foo = "baz"

    [[override]]
    when.env = "prod"
    foo = "qux"
    """

    config = tomli.loads(raw_config)
    with pytest.raises(exceptions.DuplicateError):
        combiner.build_config(config)


def test_create_outputs__dimension_not_found_in_output():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [[output]]
    region = "us"
    """

    config = tomli.loads(raw_config)
    with pytest.raises(exceptions.DimensionNotFound):
        combiner.build_config(config)


def test_create_outputs__dimension_value_not_found_in_output():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [[output]]
    env = "staging"
    """

    config = tomli.loads(raw_config)
    with pytest.raises(exceptions.DimensionValueNotFound):
        combiner.build_config(config)


def test_create_outputs__dimension_not_found_in_override():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [[output]]
    env = "dev"

    [[override]]
    when.region = "eu"
    """

    config = tomli.loads(raw_config)
    with pytest.raises(exceptions.DimensionNotFound):
        combiner.build_config(config)


def test_create_outputs__dimension_value_not_found_in_override():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [[output]]
    env = "dev"

    [[override]]
    when.env = "staging"
    """

    config = tomli.loads(raw_config)
    with pytest.raises(exceptions.DimensionValueNotFound):
        combiner.build_config(config)


def test_output_id():
    output = combiner.Output(dimensions={"env": "dev", "region": "us"}).id
    assert output == "dev-us"


@pytest.fixture
def config():
    return combiner.build_config(
        tomli.loads(
            """
        [dimensions]
        env = ["dev", "prod"]

        [default]
        foo = "bar"

        [[output]]
        env = "dev"
        """,
        )
    )


def test_generate_outputs(config):
    # Generate final configurations for each output
    assert combiner.generate_outputs(config=config) == {
        "dev": {
            "foo": "bar",
            "dimensions": {"env": "dev"},
        }
    }


def test_generate_outputs__filter_wrong_dimension(config):
    with pytest.raises(exceptions.DimensionNotFound):
        combiner.generate_outputs(
            config=config,
            foo="bar",
        )


def test_generate_outputs__filter_wrong_dimension_value(config):
    with pytest.raises(exceptions.DimensionValueNotFound):
        combiner.generate_outputs(
            config=config,
            env="foo",
        )



from __future__ import annotations

import json
import pathlib

from toml_combine import cli


def test_cli(capsys):
    """Test the CLI."""
    cli.cli(argv=["tests/test.toml"])
    out, _ = capsys.readouterr()

    expected = json.loads((pathlib.Path(__file__).parent / "result.json").read_text())
    assert json.loads(out) == expected



from __future__ import annotations

import json
import pathlib

import pytest

import toml_combine
from toml_combine import toml

config_file = pathlib.Path(__file__).parent / "test.toml"


@pytest.fixture
def expected():
    return json.loads((pathlib.Path(__file__).parent / "result.json").read_text())


@pytest.mark.parametrize(
    "kwargs",
    [
        {"config_file": config_file},
        {"config_file": str(config_file)},
        {"config": config_file.read_text()},
        {"config": toml.loads(config_file.read_text())},
    ],
)
def test_full(kwargs, expected):
    result = toml_combine.combine(**kwargs)
    assert set(result) == set(expected)
    for key in result:
        assert result[key] == expected[key], f"Failed for {key}"


# environment = ["staging", "production"]
# type = ["service", "job"]
# stack = ["next", "django"]
# service = ["api", "admin"]
# job = ["manage", "special-command"]


def test_filter__str(expected):
    result = toml_combine.combine(
        config_file=config_file, environment="production", stack="next"
    )
    assert result == {"production-service-next": expected["production-service-next"]}


def test_filter__list(expected):
    result = toml_combine.combine(
        config_file=config_file,
        environment=["production", "staging"],
        type="job",
        job=["manage", "special-command"],
    )
    assert result == {
        "staging-job-django-manage": expected["staging-job-django-manage"],
        "staging-job-django-special-command": expected[
            "staging-job-django-special-command"
        ],
        "production-job-django-manage": expected["production-job-django-manage"],
        "production-job-django-special-command": expected[
            "production-job-django-special-command"
        ],
    }



from __future__ import annotations

import copy
import dataclasses
import itertools
from collections.abc import Mapping, Sequence
from functools import partial
from typing import Any, TypeVar, Union

from . import exceptions


@dataclasses.dataclass()
class Output:
    dimensions: Mapping[str, str]

    @property
    def id(self) -> str:
        return f"{'-'.join(self.dimensions.values())}"

    def __str__(self) -> str:
        return f"Output(id={self.id})"


@dataclasses.dataclass()
class Override:
    when: Mapping[str, list[str]]
    config: Mapping[str, Any]

    def __str__(self) -> str:
        return f"Override({self.when})"


@dataclasses.dataclass()
class Config:
    dimensions: Mapping[str, list[str]]
    outputs: list[Output]
    default: Mapping[str, Any]
    overrides: Sequence[Override]


def wrap_in_list(value: str | list[str]) -> list[str]:
    """
    Wrap a string in a list if it's not already a list.
    """
    if isinstance(value, str):
        return [value]
    return value


T = TypeVar("T", bound=Union[str, list[str]])


def clean_dimensions_dict(
    to_sort: Mapping[str, T], clean: dict[str, list[str]], type: str
) -> dict[str, T]:
    """
    Recreate a dictionary of dimension values with the same order as the
    dimensions list.
    """
    result = {}
    remaining = dict(to_sort)

    for dimension, valid_values in clean.items():
        valid_values = set(valid_values)
        if dimension not in to_sort:
            continue

        original_value = remaining.pop(dimension)
        values = set(wrap_in_list(original_value))
        if invalid_values := values - valid_values:
            raise exceptions.DimensionValueNotFound(
                type=type,
                id=to_sort,
                dimension=dimension,
                value=", ".join(invalid_values),
            )
        result[dimension] = original_value

    if remaining:
        raise exceptions.DimensionNotFound(
            type=type,
            id=to_sort,
            dimension=", ".join(to_sort),
        )

    return result


def override_sort_key(
    override: Override, dimensions: dict[str, list[str]]
) -> tuple[int, ...]:
    """
    We sort overrides before applying them, and they are applied in the order of the
    sorted list, each override replacing the common values of the previous overrides.

    override_sort_key defines the sort key for overrides that ensures less specific
    overrides come first:
    - Overrides with fewer dimensions come first (will be overridden
      by more specific ones)
    - If two overrides have the same number of dimensions but define different
      dimensions, we sort by the definition order of the dimensions.

    Example:
    dimensions = {"env": ["dev", "prod"], "region": ["us", "eu"]}

    - Override with {"env": "dev"} comes before override with
      {"env": "dev", "region": "us"} (less specific)
    - Override with {"env": "dev"} comes before override with {"region": "us"} ("env"
      is defined before "region" in the dimensions list)
    """
    result = [len(override.when)]
    for i, dimension in enumerate(dimensions):
        if dimension in override.when:
            result.append(i)

    return tuple(result)


def merge_configs(a: Any, b: Any, /) -> Any:
    """
    Recursively merge two configuration dictionaries, with b taking precedence.
    """
    if isinstance(a, dict) != isinstance(b, dict):
        raise ValueError(f"Cannot merge {type(a)} with {type(b)}")

    if not isinstance(a, dict):
        return b

    result = a.copy()
    for key, b_value in b.items():
        if a_value := a.get(key):
            result[key] = merge_configs(a_value, b_value)
        else:
            result[key] = b_value
    return result


def build_config(config: dict[str, Any]) -> Config:
    # Parse dimensions
    dimensions = config.pop("dimensions")

    # Parse template
    default = config.pop("default", {})

    seen_conditions = set()
    overrides = []
    for override in config.pop("override", []):
        try:
            when = override.pop("when")
        except KeyError:
            raise exceptions.MissingOverrideCondition(id=override)

        when = {k: wrap_in_list(v) for k, v in when.items()}

        conditions = tuple((k, tuple(v)) for k, v in when.items())
        if conditions in seen_conditions:
            raise exceptions.DuplicateError(type="override", id=when)

        seen_conditions.add(conditions)

        overrides.append(
            Override(
                when=clean_dimensions_dict(
                    to_sort=when, clean=dimensions, type="override"
                ),
                config=override,
            )
        )
    # Sort overrides by increasing specificity
    overrides = sorted(
        overrides,
        key=partial(override_sort_key, dimensions=dimensions),
    )

    outputs = []
    seen_conditions = set()

    for output in config.pop("output", []):
        for key in output:
            output[key] = wrap_in_list(output[key])

        for cartesian_product in itertools.product(*output.values()):
            single_output = dict(zip(output.keys(), cartesian_product))

            conditions = tuple(single_output.items())
            if conditions in seen_conditions:
                raise exceptions.DuplicateError(type="output", id=output.id)
            seen_conditions.add(conditions)

            outputs.append(
                Output(
                    dimensions=clean_dimensions_dict(
                        single_output, dimensions, type="output"
                    ),
                )
            )

    return Config(
        dimensions=dimensions,
        outputs=outputs,
        default=default,
        overrides=overrides,
    )


def output_matches_override(output: Output, override: Override) -> bool:
    """
    Check if the values in the override match the output dimensions.
    """
    for dim, values in override.when.items():
        if dim not in output.dimensions:
            return False

        if output.dimensions[dim] not in values:
            return False

    return True


def generate_output(
    default: Mapping[str, Any], overrides: Sequence[Override], output: Output
) -> dict[str, Any]:
    result = copy.deepcopy(default)
    # Apply each matching override
    for override in overrides:
        # Check if all dimension values in the override match

        if output_matches_override(output=output, override=override):
            result = merge_configs(result, override.config)

    return {"dimensions": output.dimensions, **result}


def generate_outputs(config: Config, **filter: str | list[str]) -> dict[str, Any]:
    result = {}
    filter_with_lists: dict[str, list[str]] = {}

    for key, value in list(filter.items()):
        if key not in config.dimensions:
            raise exceptions.DimensionNotFound(type="arguments", id="", dimension=key)

        value = wrap_in_list(value)

        if set(value) - set(config.dimensions[key]):
            raise exceptions.DimensionValueNotFound(
                type="arguments",
                id="",
                dimension=key,
                value=", ".join(set(value) - set(config.dimensions[key])),
            )
        filter_with_lists[key] = value

    for output in config.outputs:
        if all(
            output.dimensions.get(key) in value
            for key, value in filter_with_lists.items()
        ):
            result[output.id] = generate_output(
                default=config.default,
                overrides=config.overrides,
                output=output,
            )

    return result



from __future__ import annotations

from typing import Any

import tomli

from . import exceptions


def loads(raw_config: str) -> dict[str, Any]:
    try:
        return tomli.loads(raw_config)
    except tomli.TOMLDecodeError as e:
        raise exceptions.TomlDecodeError(
            message="Syntax error in TOML configuration file",
            context=str(e),
        ) from e
